from datetime import date
from typing import Any, Dict, List, Literal, Optional, Union, cast

from .. import enums
from ._client_protocol import ClientProtocol


class ClassApiMixin(ClientProtocol):
    "Methods for working with classes."

    class_default_props = ["cls_id", "cls_name", "com_id"]
    class_props = [
        "cls_change_ex",
        "cls_def_pw_account",
        "cls_def_pw_console",
        "cls_def_pw_enable",
        "cls_email_logs",
        "cls_end_date",
        "cls_id",
        "cls_lab_limit",
        "cls_max_slots_per_res",
        "cls_min_hours_btw_res",
        "cls_name",
        "cls_no_delete",
        "cls_retain_ilt",
        "cls_retain_period",
        "cls_retain_st",
        "cls_self_sched",
        "cls_start_date",
        "cls_team_sched",
        "cls_uuid",
        "com_id",
        "enrollment",
        "leads",
    ]

    roster_default_props = ["acc_id", "acc_user_id", "acc_email", "lead"]
    roster_props = [
        "acc_display_name",
        "acc_email",
        "acc_full_name",
        "acc_id",
        "acc_last_login",
        "acc_logins",
        "acc_sort_name",
        "acc_time_last_access",
        "acc_type",
        "acc_user_id",
        "cls_id",
        "lead",
        "ros_team",
    ]

    async def class_list(
        self,
        *,
        com_id: Optional[int] = None,
        member: bool = False,
        properties: Union[List[str], Literal["default", "all"]] = "default",
        **kwargs,
    ) -> List[Dict[str, Any]]:
        """
        This method allows you to retrieve a list of classes and their properties.

        :param com_id: Community ID.
        :param member: If true, get list of classes the caller is in.
        :param properties: List of properties to retrieve. See Properties section for list of available properties.

        :return: On success, returns an object with the requested properties:

        **Properties**

        cls_change_ex
            Allow student to change exercise.
        cls_def_pw_account
            Class default account password.
        cls_def_pw_console
            Class default console password.
        cls_def_pw_enable
            Class default enable password.
        cls_email_logs
            Email reservation logs for real equipment pods. :py:class:`netlab.enums.ClassEmailLogs`
        cls_end_date
            Class end date.
        cls_id
            Unique class identifier.
        cls_lab_limit
            Lab limit settings. :py:class:`netlab.enums.ClassLabLimit`
        cls_max_slots_per_res
            Maximum number of slots per reservation.
        cls_min_hours_btw_res
            Minimum number of hours between reservations.
        cls_name
            Unique class name per community.
        cls_no_delete
            Prevent from deletion.
        cls_retain_ilt
            Retain instructor led logs.
        cls_retain_period
            Number of days to retain class logs.
        cls_retain_st
            Retain student logs.
        cls_self_sched
            Allow students to self schedule.
        cls_start_date
            Class start date.
        cls_team_sched
            Allow teams to schedule.
        cls_uuid
            Class UUID.
        com_id
            Unique community identifier.
        enrollment
            Count of those enrolled.
        leads
            List of leads and properties sorted by acc_sort_name.
        """
        method = "class.list"
        if properties == "all":
            properties = self.class_props
        if properties == "default":
            properties = self.class_default_props
        return await self.call(method, com_id=com_id, member=member, properties=properties, **kwargs)

    async def class_add(
        self,
        *,
        cls_name: str,
        com_id: int,
        cls_change_ex: bool = True,
        cls_def_pw_account: Optional[str] = None,
        cls_def_pw_console: Optional[str] = None,
        cls_def_pw_enable: Optional[str] = None,
        cls_email_logs: enums.ClassEmailLogs = enums.ClassEmailLogs.NO,
        cls_end_date: Optional[date] = None,
        cls_lab_limit: enums.ClassLabLimit = enums.ClassLabLimit.ENFORCE,
        cls_max_slots_per_res: Optional[int] = None,
        cls_min_hours_btw_res: Optional[int] = None,
        cls_no_delete: bool = False,
        cls_retain_ilt: bool = False,
        cls_retain_period: Optional[date] = None,
        cls_retain_st: bool = True,
        cls_self_sched: bool = True,
        cls_start_date: Optional[date] = None,
        cls_team_sched: bool = False,
        **kwargs,
    ) -> int:
        """
        This method allows you to add a class to the NETLAB+ system.

        :param cls_name: Class name.
        :param com_id: Community ID.

        :param cls_change_ex: Allow student to change exercise.
        :param cls_def_pw_account: Class default account password.
        :param cls_def_pw_console: Class default console password.
        :param cls_def_pw_enable: Class default enable password.
        :param cls_email_logs: Email reservation logs for real equipment pods.
        :param cls_end_date: Class end date.
        :param cls_lab_limit: Class lab time limit.
        :param cls_max_slots_per_res: Maximum number of slots per reservation.
        :param cls_min_hours_btw_res: Minimum hours between reservations.
        :param cls_no_delete: Prevent from deletion.
        :param cls_retain_ilt: Retain instructor led logs.
        :param cls_retain_period: The number of days to retain logs for.
        :param cls_retain_st: Retain student logs.
        :param cls_self_sched: Allow students to self schedule.
        :param cls_start_date: Class start date.
        :param cls_team_sched: Allow teams to schedule.

        :return: On success, returns the **cls_id** of the newly created class.
        """

        method = "class.add"
        params = {
            "cls_name": cls_name,
            "com_id": com_id,
            "cls_change_ex": cls_change_ex,
            "cls_email_logs": cls_email_logs,
            "cls_lab_limit": cls_lab_limit,
            "cls_no_delete": cls_no_delete,
            "cls_retain_ilt": cls_retain_ilt,
            "cls_retain_st": cls_retain_st,
            "cls_self_sched": cls_self_sched,
            "cls_team_sched": cls_team_sched,
        }

        if cls_def_pw_account:
            params["cls_def_pw_account"] = cls_def_pw_account
        if cls_def_pw_console:
            params["cls_def_pw_console"] = cls_def_pw_console
        if cls_def_pw_enable:
            params["cls_def_pw_enable"] = cls_def_pw_enable

        if cls_start_date:
            params["cls_start_date"] = cls_start_date
        if cls_end_date:
            params["cls_end_date"] = cls_end_date

        if cls_max_slots_per_res:
            params["cls_max_slots_per_res"] = cls_max_slots_per_res
        if cls_min_hours_btw_res:
            params["cls_min_hours_btw_res"] = cls_min_hours_btw_res

        if cls_retain_period:
            params["cls_retain_period"] = cls_retain_period

        params.update(kwargs)
        return int(await self.call(method, **params))

    async def class_get(
        self, *, cls_id: int, properties: Union[List[str], Literal["all", "default"]] = "default", **kwargs
    ) -> Dict[str, Any]:
        """
        This method allows you to retrieve a single class and properties.

        :param int cls_id: Class ID.
        :param properties: List of properties to retrieve. Defaults to "all".
        :return: On success, returns an object with the requested properties:

        **Properties**

        cls_change_ex
            Allow student to change exercise.
        cls_def_pw_account
            Class default account password.
        cls_def_pw_console
            Class default console password.
        cls_def_pw_enable
            Class default enable password.
        cls_email_logs
            Email reservation logs for real equipment pods. :py:class:`netlab.enums.ClassEmailLogs`
        cls_end_date
            Class end date.
        cls_id
            Unique class identifier.
        cls_lab_limit
            Lab limit settings. :py:class:`netlab.enums.ClassLabLimit`
        cls_max_slots_per_res
            Maximum number of slots per reservation.
        cls_min_hours_btw_res
            Minimum number of hours between reservations.
        cls_name
            Unique class name per community.
        cls_no_delete
            Prevent from deletion.
        cls_retain_ilt
            Retain instructor led logs.
        cls_retain_period
            Number of days to retain class logs.
        cls_retain_st
            Retain student logs.
        cls_self_sched
            Allow students to self schedule.
        cls_start_date
            Class start date.
        cls_team_sched
            Allow teams to schedule.
        cls_uuid
            Class UUID.
        com_id
            Unique community identifier.
        enrollment
            Count of those enrolled.
        leads
            List of leads and properties sorted by acc_sort_name.
        """

        method = "class.get"
        if properties == "all":
            properties = self.class_props
        if properties == "default":
            properties = self.class_default_props
        return await self.call(method, cls_id=cls_id, properties=properties, **kwargs)

    async def class_update(
        self,
        *,
        cls_id: int,
        cls_change_ex: Optional[bool] = None,
        cls_def_pw_account: Optional[str] = None,
        cls_def_pw_console: Optional[str] = None,
        cls_def_pw_enable: Optional[str] = None,
        cls_email_logs: Optional[enums.ClassEmailLogs] = None,
        cls_end_date: Optional[date] = None,
        cls_lab_limit: Optional[enums.ClassLabLimit] = None,
        cls_max_slots_per_res: Optional[int] = None,
        cls_min_hours_btw_res: Optional[int] = None,
        cls_name: Optional[str] = None,
        cls_no_delete: Optional[bool] = None,
        cls_retain_ilt: Optional[bool] = None,
        cls_retain_period: Optional[int] = None,
        cls_retain_st: Optional[bool] = None,
        cls_self_sched: Optional[bool] = None,
        cls_start_date: Optional[date] = None,
        cls_team_sched: Optional[bool] = None,
        cls_ext_slots_per_res: Optional[Union[int, enums.ClassExtensionSlots]] = None,
        **kwargs,
    ) -> Literal["OK"]:
        """
        This method allows you to update a class and its properties.

        :param cls_id: Unique class identifier.
        :param cls_change_ex: Allow student to change exercise.
        :param cls_def_pw_account: Class default account password.
        :param cls_def_pw_console: Class default console password.
        :param cls_def_pw_enable: Class default enable password.
        :param cls_email_logs: Email reservation logs for real equipment pods.
        :param cls_end_date: Class end date.
        :param cls_lab_limit: Class lab time limit.
        :param cls_max_slots_per_res: Maximum number of slots per reservation.
        :param cls_min_hours_btw_res: Minimum hours between reservations.
        :param cls_name: Class name.
        :param cls_no_delete: Prevent from deletion.
        :param cls_retain_ilt: Retain instructor led logs.
        :param cls_retain_period: The number of days to retain logs for.
        :param cls_retain_st: Retain student logs.
        :param cls_self_sched: Allow students to self schedule.
        :param cls_start_date: Class start date.
        :param cls_team_sched: Allow teams to schedule.
        :param cls_ext_slots_per_res: Allow extensions to reservations for students and teams.
            Instructors are not restricted to this limit.
        """

        method = "class.update"
        params: Dict[str, Any] = {"cls_id": cls_id}

        if cls_name:
            params["cls_name"] = cls_name
        if cls_start_date:
            params["cls_start_date"] = cls_start_date
        if cls_end_date:
            params["cls_end_date"] = cls_end_date
        if cls_def_pw_account:
            params["cls_def_pw_account"] = cls_def_pw_account
        if cls_def_pw_console:
            params["cls_def_pw_console"] = cls_def_pw_console
        if cls_def_pw_enable:
            params["cls_def_pw_enable"] = cls_def_pw_enable
        if cls_email_logs:
            params["cls_email_logs"] = cls_email_logs
        if cls_lab_limit:
            params["cls_lab_limit"] = cls_lab_limit
        if cls_no_delete is not None:
            params["cls_no_delete"] = cls_no_delete
        if cls_retain_ilt is not None:
            params["cls_retain_ilt"] = cls_retain_ilt
        if cls_retain_st is not None:
            params["cls_retain_st"] = cls_retain_st
        if cls_retain_period:
            params["cls_retain_period"] = cls_retain_period
        if cls_self_sched is not None:
            params["cls_self_sched"] = cls_self_sched
        if cls_team_sched is not None:
            params["cls_team_sched"] = cls_team_sched
        if cls_change_ex is not None:
            params["cls_change_ex"] = cls_change_ex
        if cls_min_hours_btw_res:
            params["cls_min_hours_btw_res"] = cls_min_hours_btw_res
        if cls_max_slots_per_res:
            params["cls_max_slots_per_res"] = cls_max_slots_per_res
        if cls_ext_slots_per_res is not None:
            params["cls_ext_slots_per_res"] = cls_ext_slots_per_res

        params.update(kwargs)
        return await self.call(method, **params)

    async def class_remove(self, *, cls_id: int, delete_students: bool = False, **kwargs) -> None:
        """
        This method removes a class from the NETLAB+ system.

        :param cls_id: Local system class.
        :param delete_students: `False` Do not delete student accounts. `True` Delete student accounts
            that are exclusively in this class. Student accounts that are members of multiple classes are not deleted.
        """

        method = "class.remove"
        return await self.call(method, cls_id=cls_id, delete_students=delete_students, **kwargs)

    async def class_roster_add(
        self, *, acc_id: str, cls_id: str, lead: bool = False, ros_team: Optional[str] = None, **kwargs
    ) -> Literal["OK"]:
        """
        This method allows you to add a member to a class roster.

        :param acc_id: Unique account identifier.
        :param cls_id: Unique class identifier.
        :param lead: Add as lead.
        :param ros_team: Assign user to a team. Single character, A-Z.
        """

        method = "class.roster.add"
        params = {"acc_id": acc_id, "cls_id": cls_id, "lead": lead}

        if ros_team:
            params["ros_team"] = ros_team

        params.update(kwargs)
        return await self.call(method, **params)

    async def class_roster_list(self, *, cls_id: int, leads: bool = False, **kwargs):
        """
        This method lists users in the class roster.

        :param cls_id: Local system class.
        :param leads: `None` Return all users in the roster (leads and learners).
            `True` Return lead instructors only. `False` Return learners only.

        :return: On success, returns an object with the requested properties:

        **Properties**

        acc_display_name
            Account name people will see this name in discussions, messages and comments.
        acc_email
            Account email.
        acc_id
            Local system account identifier.
        acc_full_name
            Account full name.
        acc_last_login
            Timestamp of last account login. Possibly `None`.
        acc_logins
            Number of times a user has logged in.
        acc_sort_name
            Account name as it appears in sorted lists.
        acc_time_last_access
            Timestamp of last access. Possibly `None`.
        acc_type
            Account type. :py:class:`netlab.enums.AccountType`
        acc_user_id
            Account user identifier.
        cls_id
            Unique class identifier.
        lead
            If user is a lead instructor.
        ros_team
            Assignment of user to a team (designated by team letter). `None` if user is not assigned to a team.
        """

        method = "class.roster.list"
        return await self.call(method, cls_id=cls_id, leads=leads, **kwargs)

    async def class_roster_remove(self, *, acc_id: int, cls_id: str, **kwargs) -> None:
        """
        This method removes a user from the class roster.

        :param acc_id: Account identifier of user to be removed.
        :param cls_id: Unique class identifier.
        """

        method = "class.roster.remove"
        return await self.call(method, acc_id=acc_id, cls_id=cls_id, **kwargs)

    async def class_roster_get(
        self,
        *,
        cls_id: int,
        acc_id: Optional[int],
        **kwargs,
    ) -> Dict[str, Any]:
        """
        This method returns class roster information for a single user.

        :param cls_id: Local system class identifier.
        :param acc_id: Returns roster information for the user specified by acount ID acc_id.
            If acc_id is not provided, roster information related to the requester is returned.
        :return: On success, returns an object with the requested properties:

        **Properties**

        acc_display_name
            Account name people will see this name in discussions, messages and comments.
        acc_email
            Account email.
        acc_id
            Local system account identifier.
        acc_full_name
            Account full name.
        acc_last_login
            Timestamp of last account login.
        acc_logins
            Number of times a user has logged in.
        acc_sort_name
            Account name as it appears in sorted lists.
        acc_time_last_access
            Timestamp of last access.
        acc_type
            Account type.
        acc_user_id
            Account user identifier.
        cls_id
            Unique class identifier.
        lead
            User is a lead instructor.
        ros_team
            Team identifier [A-Z] or `None` if user is not assigned to a team.
        """
        method = "class.roster.get"

        return await self.call(method, acc_id=acc_id, cls_id=cls_id, **kwargs)

    async def class_roster_team_update(
        self,
        *,
        cls_id: int,
        ros_team: Optional[str] = None,
        roster_acc_id: List[int],
        **kwargs,
    ) -> str:
        """
        This method will create and update teams (groups of students) for a given class roster.

        :param cls_id: Local system class identifier.
        :param ros_team: A single uppercase letter [A,Z] as the designation of the team to add
            the student(s) to. If this value is `None`, the student(s) are assigned to no team
            (unassigned).
        :param roster_acc_id: A list of account IDs to assign to the specified team.
        """

        method = "class.roster.team.update"

        acc_id_csv = ",".join(map(str, roster_acc_id))

        return await self.call(method, cls_id=cls_id, ros_team=ros_team, roster_acc_id=acc_id_csv, **kwargs)

    async def class_content_add(
        self,
        *,
        cls_id: int,
        con_id: str,
        **kwargs,
    ) -> None:
        """
        This method allows you to add content to a class.

        :param cls_id: Class identifier.
        :param con_id: Content identifier.
        """

        method = "class.content.add"
        return await self.call(
            method,
            cls_id=cls_id,
            con_id=con_id,
            **kwargs,
        )

    async def class_content_availability(
        self,
        *,
        cls_id: int,
        **kwargs,
    ) -> List[Dict[str, Any]]:
        """
        This method provides a list of available class content.

        :param cls_id: Unnique class identifier.

        :return: On success, returns a list of objects with the requested properties:

        **Properties**

        con_access
            Content is Globally accessible or Private.
        con_aid
            Content author ID.
        con_author
            Content author name.
        con_build
            Build number.
        con_copyright
            Content copyright.
        con_desc
            Content description.
        con_gid
            Unique content identifier.
        con_global
            Is content globally accessible or not.
        con_id
            Unique content identifier.
        con_managed
            Is the content managed or not.
        con_name
            Name of content.
        con_notes
            Content notes.
        con_org
            Content organization.
        con_origin
            Source of the content.
        con_pt_ids
            Comma separated list of required Pod Types.
        con_url
            Content URL.
        """
        method = "class.content.availability"

        return await self.call(
            method,
            cls_id=cls_id,
            **kwargs,
        )

    async def class_content_list(
        self,
        *,
        cls_id: int,
        properties: Union[Literal["all"], List[str]] = "all",
        **kwargs,
    ) -> List[Dict[str, Any]]:
        """
        This method provides a list of content associated with a class.

        :param cls_id: Unique class identifier.
        :param properties: Array of properties to retrieve.

        :return: On success, returns a list of objects with the requested properties:

        **Properties**

        con_access
            Content is Globally accessible or Private.
        con_aid
            Content author ID.
        con_author
            Content author name.
        con_build
            Build number.
        con_copyright
            Content copyright.
        con_desc
            Content description.
        con_gid
            Unique content identifier.
        con_global
            Is content globally accessible or not.
        con_id
            Unique content identifier.
        con_managed
            Is the content managed or not.
        con_name
            Name of content.
        con_notes
            Content notes.
        con_org
            Content organization.
        con_origin
            Source of the content.
        con_pt_ids
            Comma separated list of required Pod Types.
        con_url
            Content URL.
        """
        method = "class.content.list"

        if properties == "all":
            properties = cast(Literal["all"], None)

        return await self.call(
            method,
            cls_id=cls_id,
            properties=properties,
            **kwargs,
        )

    async def class_content_remove(
        self,
        *,
        cls_id: int,
        con_id: str,
        **kwargs,
    ) -> None:
        """
        This method removes content from a class.

        :param cls_id: Class identifier.
        :param con_id: Content identifier.
        """
        method = "class.content.remove"

        return await self.call(
            method,
            cls_id=cls_id,
            con_id=con_id,
            **kwargs,
        )



from dataclasses import dataclass
from typing import NamedTuple

from ..enums import VirtualMachineRole


class NetlabHost(NamedTuple):
    vh_id: int
    vh_name: str

    def __str__(self) -> str:
        return f"{self.vh_name}/{self.vh_id}"


@dataclass
class ExistingPod:
    pod_id: int
    current_name: str
    current_role: VirtualMachineRole | None
    current_host: NetlabHost | None
    current_pod_type: str

    pt_apdid: str
    pt_org: str

    def get_host(self) -> NetlabHost | None:
        return self.current_host

    def get_name(self) -> str:
        return self.current_name

    def get_role(self) -> VirtualMachineRole | None:
        return self.current_role

    def get_pod_type(self) -> str:
        return self.current_pod_type



import csv
import os
import sys
from typing import Any

from netlab.client import Client
from netlab.errors.common import NetlabError

#####################
# Variables to change
#####################

COMMUNITY_ID = 1  # User accounts must belong to a community.
CLASS_ID = None  # All users will be added directly to this class during account creation.
# Set CLASS_ID to None if you don't want to enroll users into a class automatically.
DEFAULT_INITIAL_PASSWORD = "CHANGEMEPLEASE"  # can be overidden per user by including acc_password in the CSV
NETLAB_SYSTEM = "DEMO"  # must be a valid entry in your config.json

#################################################################
# Don't change the code below unless you know what you are doing.
#################################################################

try:
    # Pass a filename as the first argument to the script
    user_csv_file = sys.argv[1]
except IndexError:
    # Or default to users.csv
    user_csv_file = "users.csv"

print("Opening {}".format(user_csv_file))

if not os.path.exists(user_csv_file):
    print("    File not found!")
    sys.exit(1)
else:
    print("  -> Beginning user import")

client = Client(NETLAB_SYSTEM)
users_created = 0
users_with_errors = 0

with open(user_csv_file) as csvfile:
    reader = csv.DictReader(csvfile)
    row: Any
    for row in reader:
        # Use the default password if acc_password is not a column in the csv or if it is empty
        if not row.get("acc_password", ""):
            row["acc_password"] = DEFAULT_INITIAL_PASSWORD

        try:
            acc_id = client.user_account_add(com_id=COMMUNITY_ID, cls_id=CLASS_ID, **row)
        except NetlabError as err:
            # The server returned an error. The exception will contain more information about the error.
            users_with_errors += 1
            print("  -! {}".format(err))
        else:
            users_created += 1
            print("  -+ Created user {} with id {}".format(row["acc_user_id"], acc_id))

print("\nSuccessfully imported {} new users.".format(users_created))
if users_with_errors > 0:
    print("There were {} errors encountered. Please review the output above.".format(users_with_errors))



import asyncio
import platform
from importlib.metadata import version as get_version
from typing import Any

import click

from ..async_client import NetlabClient
from .api import api
from .build import build
from .config import config
from .debug import debug
from .utils import print_exception

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.group(context_settings=CONTEXT_SETTINGS)
def main():
    pass


@main.command()
@click.option(
    "--system", default="default", help="The NETLAB+ system to get server version from. Leave blank for 'default'."
)
def version(system):
    """Show the client, server, Python, and OS versions"""

    asyncio.run(_async_version(system))


async def _async_version(system):
    results: Any
    try:
        async with NetlabClient(system) as client:
            results = await client.system_status_get()
    except Exception as e:
        results = {}
        print_exception(e)
        click.echo("")
    server_version = results.get("sys_sdn_version", click.style("Unable to connect to server", fg="red"))

    click.echo("NETLAB+ Server : {}".format(server_version))
    click.echo("Python Client  : {}".format(get_version("netlab")))
    click.echo("Python Version : {}".format(platform.python_version()))
    click.echo("OS Platform    : {}".format(platform.platform()))


# manually setup top level command groups
main.add_command(api)
main.add_command(config)
main.add_command(debug)
main.add_command(build)



from typing import Literal, Optional

from typing_extensions import TypedDict  # Needed for pydantic (<py3.11)


class PCCloneSpec(TypedDict, total=False):
    "Instructions for a clone."

    clone_datastore: str
    "The name of the datastore the cloned VM will reside."
    clone_name: str
    "Cloned VM name. Default: {clone_pod_name}_{pc_label}"
    clone_role: Literal["MASTER", "NORMAL", "PERSISTENT", "TEMPLATE"]
    """
    :MASTER: Commonly used for master pod deployment. Can not be easily deleted.
    :NORMAL: *Default* Standard virtual machine image for use in pods.
    :PERSISTENT: Commonly used when the virtual machine is not reverting to snapshot and is perpetual.
    :TEMPLATE: Pristine virtual machine image used as the basis for cloning many virtual machines.
        Template VMs cannot be powered on, modified, or assigned to pods.
    """
    clone_snapshot: Optional[str]
    """Name of the snapshot to take on cloned VM. "GOLDEN" is recommended. If set to NONE, a
    snapshot will not be taken of the cloned VM. If absent, "GOLDEN" will be used."""
    clone_storage_alloc: Literal["ONDEMAND", "PREALLOCATED"]
    """
    :ONDEMAND: *Default* Actual size of disk will grow as data is added (saves on storage).
    :PREALLOCATED: The entire disk size will be allocated immediately.
    """
    clone_type: Literal["LINKED", "FULL"]
    """
    :LINKED: *Default* Linked clone is a delta of a full virtual machine.
    :FULL: Full clone is a complete copy of the virtual machine.
    """
    clone_vh_id: int
    "The runtime host of cloned VM."
    copy_bios_uuid: bool
    "If true, copy this VMs BIOS UUID to any VM cloned from this one."
    pc_type: Literal["AVMI", "ABSENT"]
    """
    :AVMI: Use virtual machine inventory.
    :ABSENT: PC is absent.
    """
    pl_index: int
    "PC index of the pod design."
    source_snapshot: str
    "Source VM snapshot name. Default: vm_snapshot"
    source_vm_id: str
    "Source VM pod identifier."



from enum import Enum


class AccountPrivileges(Enum):
    ""

    COMWIDE = "COMWIDE"
    "Community Administrator."
    SYSWIDE = "SYSWIDE"
    "Systemwide Administrator."
    POD_DESIGNER = "POD_DESIGNER"
    "Allows use of Pod Designer."
    LAB_DESIGNER = "LAB_DESIGNER"
    "Allows use of Lab Designer."


class AccountType(Enum):
    """
    Account type.
    """

    STUDENT = "S"
    "Student"
    INSTRUCTOR = "I"
    "Instructor"
    ADMIN = "Z"
    "Admin"


class CommunityExtensionSlots(Enum):
    "The number of times a reservation can be extended."

    UNLIMITED = -1
    "Allow unlimited extensions."



import sys

from rich.progress import Progress, TaskID

if sys.version_info >= (3, 11):
    from asyncio import TaskGroup
else:
    from taskgroup import TaskGroup

from ..async_client import NetlabConnection
from . import schema
from .types import ExistingPod, NetlabHost


async def _fetch_hosts(client: NetlabConnection) -> dict[int | str, NetlabHost]:
    host_map: dict[int | str, NetlabHost] = {}

    for host_raw in await client.vm_host_list():
        host = NetlabHost(
            host_raw["vh_id"],
            host_raw["vh_name"],
        )
        host_map[host.vh_id] = host
        host_map[host.vh_name] = host

    return host_map


async def _fetch_pods(client: NetlabConnection, manifest: schema.Manifest) -> dict[int, ExistingPod]:
    result: dict[int, ExistingPod] = {}
    existing_pod_ids = set(await client.pod_list_used_ids())
    created_pod_ids: set[int] = set()
    source_pod_ids = set()
    for r in manifest.range:
        if r.delete:
            continue
        rng = set(r.get_range())
        source_pod_ids.add(r.source_pod_id)
        assert rng.isdisjoint(created_pod_ids)
        created_pod_ids.update(rng)

    async def inner(pod_id: int, progress_task: TaskID, progress: Progress) -> None:
        try:
            raw_pod = await client.pod_get(pod_id=pod_id, properties="all")
            # progress.console.print(f"Fetching pod {pod_id}")

            try:
                host = NetlabHost(
                    raw_pod["remote_pc"][0]["vh_id"],
                    raw_pod["remote_pc"][0]["vh_name"],
                )
                vm_role = raw_pod["remote_pc"][0]["vm_role"]
            except KeyError:
                host = None
                vm_role = None

            existing_pod = ExistingPod(
                pod_id=pod_id,
                current_host=host,
                current_role=vm_role,
                current_name=raw_pod["pod_name"],
                current_pod_type=raw_pod["pt_id"],
                pt_org=raw_pod["pt_org"],
                pt_apdid=raw_pod["pt_apdid"],
            )
            result[pod_id] = existing_pod
            progress.update(progress_task, advance=1)
        except Exception as e:
            if sys.version_info >= (3, 11):
                e.add_note(f"Fetching pod_id={pod_id}")
            raise e

    throt = manifest.throttles.fetch.make()
    pod_ids = existing_pod_ids.intersection(created_pod_ids | source_pod_ids)
    with Progress() as progress:
        progress_task = progress.add_task("Fetching pods", total=len(pod_ids))

        async with TaskGroup() as tg:
            for pod_id in pod_ids:
                bucket = await throt.bucket()
                tg.create_task(bucket(inner(pod_id, progress_task, progress)))

    return result



import asyncio
from contextlib import AbstractAsyncContextManager
from datetime import datetime, timedelta
from logging import getLogger
from typing import Any, Awaitable, TypeVar, Union

from typing_extensions import Never, deprecated, override

_background_tasks: set[asyncio.Task[Any]] = set()

log = getLogger(__name__)


class _SleepUntil(object):
    def __init__(self, dt: datetime | timedelta):
        self._dt: datetime = datetime.fromtimestamp(0)
        self.update(dt)
        self._lock = asyncio.Lock()

    def update(self, dt: datetime | timedelta) -> None:
        if isinstance(dt, timedelta):
            dt = datetime.now() + dt

        if dt > self._dt:
            self._dt = dt

    async def __call__(self) -> None:
        async with self._lock:
            while (delta := self._dt - datetime.now()) > timedelta(seconds=0):
                await asyncio.sleep(delta.total_seconds())


T = TypeVar("T")


class Throttler(object):
    """
    BETA: Provides throttling using the Leaky Bucket Algorithm.

    https://en.wikipedia.org/wiki/Leaky_bucket

    It can be imagined like so:

    There are a number of buckets, full of water. There is a tap, that can fill
    a bucket in (refresh) seconds. For a task to start, it must flip a filled
    bucket. It will keep the bucket overturned until it completes. When it
    completes, the bucket will wait to be filled so that another task can use
    it.
    """

    def __init__(
        self,
        buckets: int,
        refresh: Union[float, timedelta],
        stagger: Union[float, timedelta] = timedelta(0.0),
    ):
        """
        Create a throttle object.

        :param buckets: Number of buckets (concurrent tasks).
        :param refresh: Time (seconds) taken to fill a bucket.
        :param stagger: Stagger task start times.
        """
        if isinstance(refresh, timedelta):
            self._restore_time = refresh
        else:
            self._restore_time = timedelta(seconds=refresh)

        if isinstance(stagger, timedelta):
            self._stagger = stagger
        else:
            self._stagger = timedelta(seconds=stagger)

        self._stagger_wait = _SleepUntil(timedelta(seconds=0))

        self._filler_task = asyncio.create_task(self._filler())
        _background_tasks.add(self._filler_task)  # For GC
        self._filler_task.add_done_callback(_background_tasks.discard)

        self._buckets_total = buckets
        self._full_buckets = asyncio.Queue[None]()
        for i in range(self._buckets_total):
            self._full_buckets.put_nowait(None)

        self._empty_buckets = asyncio.Queue[None]()

    def __del__(self) -> None:
        self._filler_task.cancel()

    async def _filler(self) -> Never:
        while True:
            await self._empty_buckets.get()
            await asyncio.sleep(self._restore_time.total_seconds())
            self._full_buckets.put_nowait(None)

    async def bucket(self) -> "Bucket":
        await self._stagger_wait()
        await self._full_buckets.get()
        self._stagger_wait.update(self._stagger)

        return Bucket(self)

    @deprecated("Use .bucket")
    async def wrap(self, co: Awaitable[T]) -> T:
        async with self:
            return await co

    @deprecated("Use .bucket")
    async def __aenter__(self) -> None:
        await self._full_buckets.get()

    @deprecated("Use .bucket")
    async def __aexit__(self, exc_type, exc_value, traceback):
        self._empty_buckets.put_nowait(None)


class Bucket(AbstractAsyncContextManager[None]):
    def __init__(self, throttler: Throttler):
        self._used: bool = False
        self._throttler: Throttler = throttler

    async def __call__(self, co: Awaitable[T]) -> T:
        async with self:
            return await co

    @override
    async def __aenter__(self) -> None:
        assert self._used is False

    @override
    async def __aexit__(self, exc_type, exc_value, traceback):
        self._throttler._empty_buckets.put_nowait(None)
        self._used = True

    def __del__(self) -> None:
        if self._used is False:
            log.warn("Deleting unused bucket")



from netlab.client import Client
from netlab.enums import VirtualMachineRole

client = Client(system="stan")


datacenters = client.vm_datacenter_list()

for datacenter in datacenters:
    print(f"""
    vdc_id:   {datacenter["vdc_id"]}
    vdc_name: {datacenter["vdc_name"]}

    """)

vdc_id = int(input("Enter datacenter vdc_id: ").strip())

for host in client.vm_host_list(vdc_id=vdc_id):
    print(f"""
    vh_id:  {host["vh_id"]}
    vh_name:  {host["vh_name"]}

    """)

vh_id = int(input("Enter host vh_id: ").strip())

children_to_remove = []
masters_to_remove = []

for vm in client.vm_inventory_list(vdc_id=vdc_id):
    if vm["vh_id"] == vh_id:
        if vm["vm_role"] is not VirtualMachineRole.MASTER:
            children_to_remove.append(vm["vm_id"])
        if vm["vm_role"] is VirtualMachineRole.MASTER:
            masters_to_remove.append(vm["vm_id"])

print(f"""
preparing to delete')
  total:    {len(masters_to_remove) + len(children_to_remove)}
  masters:  {len(masters_to_remove)}
  children: {len(children_to_remove)}
""")

for vm_id in children_to_remove:
    print(f"child:  {vm_id}")

for vm_id in masters_to_remove:
    print(f"master: {vm_id}")

print("\n\n")

confirmed = input("Continue?  [N/y]").lower() == "y"

if confirmed:
    for vm_id in children_to_remove + masters_to_remove:
        print(f"Deleting {vm_id}")
        client.vm_inventory_remove_local(vm_id=vm["vm_id"])
